/**
 * Copyright (c) 2003 Billy Biggs <vektor@dumbterm.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/**
 * Nearest neighbour integration-approximating image resampler.
 *
 * Doesn't that sound cooool?
 *
 * Uses 256 samples per pixel to resize an image.  Somewhat more
 * brute-force than necessary, but speedy enough to be useful.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include "pnginput.h"
#include "pngoutput.h"

const double epsilon = 0.000001;

/* The supersampling factor. */
const int ss_factor = 16;

static double intensity_to_voltage_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    /* sRGB uses an exponent of 2.4 in its nonstandard curve, but it
     * still advertises a display gamma of 2.2.
     */
    if( val <= 0.0031308 ) {
        return 12.92 * val;
    }
    return ( ( 1.055 * pow( val, 1.0 / 2.4 ) ) - 0.055 );
}

static double voltage_to_intensity_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    if( val <= 0.04045 ) {
        return val / 12.92;
    }
    return pow( ( val + 0.055 ) / 1.055, 2.4 );
}

static double point_sample( double *image, double x, double y,
                            int width, int height, int component, int n )
{
    int xpos = (int) ((((double) width) * x) + epsilon);
    int ypos = (int) ((((double) height) * y) + epsilon);
    return image[ (ypos * width * n) + (xpos * n) + component ];
}

int main( int argc, char **argv )
{
    pngoutput_t *pngout;
    pnginput_t *pngin;
    uint8_t *scanline;
    int width, height, owidth, oheight, pixelsize;
    double *inputimg;
    int y;

    if( argc < 5 ) {
        fprintf( stderr, "usage: %s input.png output.png width height\n",
                 argv[ 0 ] );
        return 1;
    }

    pngin = pnginput_new( argv[ 1 ] );
    if( !pngin ) {
        fprintf( stderr, "nnsample: Can't open %s.\n", argv[ 1 ] );
        return 1;
    }
    width = pnginput_get_width( pngin );
    height = pnginput_get_height( pngin );
    inputimg = malloc( width * height * 4 * sizeof( double) );

    owidth = atoi( argv[ 3 ] );
    oheight = atoi( argv[ 4 ] );
    pixelsize = 3 + pnginput_has_alpha( pngin );
    scanline = malloc( owidth * 4 );

    if( !scanline || !inputimg ) {
        fprintf( stderr, "nnsample: Can't allocate memory.\n" );
        pnginput_delete( pngin );
        return 1;
    }

    pngout = pngoutput_new( argv[ 2 ], owidth, oheight,
                            pnginput_has_alpha( pngin ), 0.45, 1 );
    if( !pngout ) {
        pnginput_delete( pngin );
        fprintf( stderr, "nnsample: Can't open %s.\n", argv[ 2 ] );
        return 1;
    }

    /* Read in the entire input file. */
    for( y = 0; y < height; y++ ) {
        uint8_t *input = pnginput_get_scanline( pngin, y );
        double *output = inputimg + (y * width * 4);
        int x;

        for( x = 0; x < width; x++ ) {
            double r, g, b, a;
            double rp, gp, bp;

            if( pnginput_has_alpha( pngin ) ) {
                rp = ((double) input[ (x * 4) + 0 ]) / 255.0;
                gp = ((double) input[ (x * 4) + 1 ]) / 255.0;
                bp = ((double) input[ (x * 4) + 2 ]) / 255.0;
                a  = ((double) input[ (x * 4) + 3 ]) / 255.0;
            } else {
                rp = ((double) input[ (x * 3) + 0 ]) / 255.0;
                gp = ((double) input[ (x * 3) + 1 ]) / 255.0;
                bp = ((double) input[ (x * 3) + 2 ]) / 255.0;
                a = 1.0;
            }
            r = voltage_to_intensity_srgb( rp );
            g = voltage_to_intensity_srgb( gp );
            b = voltage_to_intensity_srgb( bp );
            output[ (x * 4) + 0 ] = r;
            output[ (x * 4) + 1 ] = g;
            output[ (x * 4) + 2 ] = b;
            output[ (x * 4) + 3 ] = a;
        }
    }
    pnginput_delete( pngin );

    /* Scale image. */
    for( y = 0; y < oheight; y++ ) {
        int x;

        for( x = 0; x < owidth; x++ ) {
            double r, g, b, a, rp, gp, bp;
            int i, j;

            r = g = b = a = 0.0;

            /* Uniformly supersample (ss_factor * ss_factor) samples. */
            for( i = 0; i < ss_factor; i++ ) {
                for( j = 0; j < ss_factor; j++ ) {
                    double xpos = ((double) (x*ss_factor) + i)
                                / ((double) (owidth * ss_factor));
                    double ypos = ((double) (y*ss_factor) + j)
                                / ((double) (oheight * ss_factor));
                    double alpha = point_sample( inputimg, xpos, ypos,
                                                 width, height, 3, 4 );

                    r += (point_sample( inputimg, xpos, ypos,
                                        width, height, 0, 4 )) * alpha;
                    g += (point_sample( inputimg, xpos, ypos,
                                        width, height, 1, 4 ) * alpha);
                    b += (point_sample( inputimg, xpos, ypos,
                                        width, height, 2, 4 ) * alpha);
                    a += alpha;
                }
            }
            a /= (double) (ss_factor * ss_factor);
            if( a < epsilon ) {
                r = g = b = a = 0.0;
            } else {
                r /= ((double) (ss_factor * ss_factor) * a);
                g /= ((double) (ss_factor * ss_factor) * a);
                b /= ((double) (ss_factor * ss_factor) * a);
            }
            rp = intensity_to_voltage_srgb( r );
            gp = intensity_to_voltage_srgb( g );
            bp = intensity_to_voltage_srgb( b );
            scanline[ (x * pixelsize) + 0 ] = (int) ((rp * 255.0) + 0.5);
            scanline[ (x * pixelsize) + 1 ] = (int) ((gp * 255.0) + 0.5);
            scanline[ (x * pixelsize) + 2 ] = (int) ((bp * 255.0) + 0.5);
            if( pnginput_has_alpha( pngin ) ) {
                scanline[ (x * pixelsize) + 3 ] = (int) ((a * 255.0) + 0.5);
            }
        }

        pngoutput_scanline( pngout, scanline );
    }

    pngoutput_delete( pngout );
    return 0;
}

